unit uMain;
{$INCLUDE colbuild.inc}

interface

uses
  SysUtils, Classes, Windows, RegExpr,
  janXMLParser2,
  uCollectionDef, uFileList, uProcess, uFunc, uConfig, uStrings;

type
  TFileType = (ftUnknown, ftLRF, ftPDF, ftRTF, ftTXT, ftEPUB);

  ECollectionBuilder = class(Exception);

  TCollectionBuilder = class
    private
      fCache: TjanXMLParser2;
      fCollections: TCollectionList;
      fFiles: TFileList;
      fConfig: TConfig;
    protected
      function LinuxPath(const Path: string): string;
      function GetNextID(Node: TjanXMLNode2): integer;
      function NextID: integer;
      function GetCacheRecords: TjanXMLNode2;
      function GetTextTag: string;
      function GetPlaylistTag: string;
      function GetPlaylistItemTag: string;
      function DateTimeToXml(DT: TDateTime): string;
      function FileNameToFileType(const FN: string): TFileType;
      function FileNameToMimeType(const FN: string): string;
      function FileTypeToExtension(const FT: TFileType): string;
      function GetMetadata(const FN: string; out Title, TitleSort, Author, AuthorSort: string): boolean;
      procedure BuildConnections_FileList_Cache;
      procedure BuildConnections_Collection_Cache;
      procedure BuildConnections_Collection_FileList;
      procedure BuildConnections; virtual;
      procedure CollectionOnConfig(Sender: TCollectionList; const Name, Value: string);
      property Cache: TjanXMLParser2 read fCache;
      property Collections: TCollectionList read fCollections;
      property Files: TFileList read fFiles;
    public
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Execute; virtual;
      property Config: TConfig read fConfig;
    end;

implementation

const Extensions : array[TFileType] of string
                 = ('', '.lrf', '.pdf', '.rtf', '.txt', '.epub');

const MimeTypes: array[TFileType] of string
               = ('', 'application/x-sony-bbeb', 'application/pdf', 'application/rtf', 'text/plain', 'application/epub');

{$IFDEF ALLOW_NONEXISTENT_CACHE}
const DEFAULT_XML_INTERNAL = '<?xml version="1.0" encoding="UTF-8"?>' + EOLN +
                             '<xdbLite xmlns="http://xscool.net/xdb/1" xmlns:xs1="http://www.kinoma.com/FskCache/1" default="string" nextID="2">' + EOLN +
                             TAB + '<indexes>' + EOLN +
                             TAB +TAB + '<index name="indexSource" type="string,unique"/>' + EOLN +
                             TAB +TAB + '<index name="indexArtist" type="string"/>' + EOLN +
                             TAB +TAB + '<index name="indexAlbum" type="string"/>' + EOLN +
                             TAB +TAB + '<index name="indexDate" type="number"/>' + EOLN +
                             TAB +TAB + '<index name="indexDateTime" type="number"/>' + EOLN +
                             TAB +TAB + '<index name="indexDimensions" type="integer"/>' + EOLN +
                             TAB +TAB + '<index name="indexDuration" type="integer"/>' + EOLN +
                             TAB +TAB + '<index name="indexGenre" type="string"/>' + EOLN +
                             TAB +TAB + '<index name="indexTitle" type="string"/>' + EOLN +
                             TAB +TAB + '<index name="indexPlaylist" type="string"/>' + EOLN +
                             TAB +TAB + '<index name="indexSize" type="integer"/>' + EOLN +
                             TAB +TAB + '<index name="indexPath" type="string,sensitive"/>' + EOLN +
                             TAB + '</indexes>' + EOLN +
                             TAB + '<records>' + EOLN +
                             TAB +TAB + '<xs1:library id="0" name="CCCCCCCCCCCC-CCCC-CCCC-CCCCCCCCCCCC" title="Library">' + EOLN +
                             TAB +TAB +TAB + '<xs1:accepted>true</xs1:accepted>' + EOLN +
                             TAB +TAB + '</xs1:library>' + EOLN +
                             TAB +TAB + '<xs1:watchSpecial id="1" name="mediaPath" title="kbook">' + EOLN +
                             TAB +TAB +TAB + '<xs1:accepted>true</xs1:accepted>' + EOLN +
                             TAB +TAB + '</xs1:watchSpecial>' + EOLN +
                             TAB + '</records>' + EOLN +
                             '</xdbLite>' + EOLN;

      DEFAULT_XML_SD = '<?xml version="1.0" encoding="UTF-8"?>' + EOLN +
                       '<cache xmlns="http://www.kinoma.com/FskCache/1">' + EOLN +
                       '</cache>' + EOLN;
{$ENDIF}

{ TCollectionBuilder }

constructor TCollectionBuilder.Create;
var Ini: string;
begin
  fCache := TjanXMLParser2.Create;
  fCollections := TCollectionList.Create;
  fCollections.OwnsObjects := True;
  fCollections.OnConfig := CollectionOnConfig;
  fFiles := TFileList.Create;
  fFiles.OwnsObjects := True;
  fConfig := TConfig.Create;
  Ini := ChangeFileExt(ParamStr(0), '.ini');
  if FileExists(Ini) then
    fConfig.LoadFromIniFile(Ini, 'Config');
end;

destructor TCollectionBuilder.Destroy;
begin
  FreeAndNil(fCache);
  FreeAndNil(fCollections);
  FreeAndNil(fFiles);
  inherited;
end;

procedure TCollectionBuilder.CollectionOnConfig(Sender: TCollectionList; const Name, Value: string);
begin
  fConfig.LoadFromNameValue(Name, Value);
end;

function TCollectionBuilder.LinuxPath(const Path: string): string;
begin
  Result := StringReplace(Path, '\', '/', [rfReplaceAll]);
end;

function TCollectionBuilder.GetNextID(Node: TjanXMLNode2): integer;
var i, j: integer;
begin
  Result := StrToIntDef(Node.AsText['id'], 0);
  for i := 0 to Pred(Node.Nodes.Count) do
    begin
    j := GetNextID(Node.Nodes[i]);
    if j > Result then
      Result := j;
    end;
end;

function TCollectionBuilder.NextID: integer;
begin
  Result := Succ(GetNextID(Cache));
end;

function TCollectionBuilder.GetCacheRecords: TjanXMLNode2;
var i: integer;
begin
  if Config.MemoryType = mtInternal then
    begin
    Result := nil;
    for i := 0 to Pred(Cache.nodes.Count) do
      if Cache.childNode[i].name = 'records' then
        begin
        Result := Cache.childNode[i];
        Break;
        end;
    end
  else
    Result := Cache;
end;

function TCollectionBuilder.GetTextTag: string;
begin
  if Config.MemoryType = mtInternal then
    Result := 'xs1:text'
  else
    Result := 'text';
end;

function TCollectionBuilder.GetPlaylistTag: string;
begin
  if Config.MemoryType = mtInternal then
    Result := 'xs1:playlist'
  else
    Result := 'playlist';
end;

function TCollectionBuilder.GetPlaylistItemTag: string;
begin
  if Config.MemoryType = mtInternal then
    Result := 'xs1:item'
  else
    Result := 'item';
end;

function TCollectionBuilder.DateTimeToXml(DT: TDateTime): string;
const Days: array[1..7] of string = ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
      Months: array[1..12] of string = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
var D, M, Y, H, Mn, S, MS: Word;
begin
  DecodeDate(DT, Y, M, D);
  DecodeTime(DT, H, Mn, S, MS);
  Result := Format('%s, %02.2d %s %04.4d %02.2d:%02.2d:%02.2d GMT', [Days[DayOfWeek(DT)], D, Months[M], Y, H, Mn, S]);
end;

function TCollectionBuilder.FileNameToFileType(const FN: string): TFileType;
var Ext: string;
begin
  Ext := ExtractFileExt(FN);
  Result := High(TFileType);
  while Result > Low(TFileType) do
    if AnsiCompareText(Ext, Extensions[Result]) = 0 then
      Break
    else
      Dec(Result);
end;

function TCollectionBuilder.FileNameToMimeType(const FN: string): string;
begin
  Result := MimeTypes[FileNameToFileType(FN)];
end;

function TCollectionBuilder.FileTypeToExtension(const FT: TFileType): string;
begin
  Result := Extensions[FT];
end;

procedure TCollectionBuilder.BuildConnections;
begin
  BuildConnections_FileList_Cache;
  BuildConnections_Collection_FileList;
  if Config.CollectionRegexp <> '' then
    Collections.BuildFromRegExp(Files, Config.CollectionRegexp, Config.CollectionRegexpTitle);
  BuildConnections_Collection_Cache;
end;

procedure TCollectionBuilder.BuildConnections_Collection_FileList;
var Col: TCollectionDef;
    Item: TCollectionItem;
    i, j, k: integer;
    s: string;
begin
  for i := 0 to Pred(Collections.Count) do
    begin
    Col := Collections[i];
    for j := 0 to Pred(Col.Count) do
      begin
      Item := Col[j];
      Item.FileNode := nil;
      s := LinuxPath(Item.Name);
      for k := 0 to Pred(Files.Count) do
        if AnsiCompareText(s, LinuxPath(Files[k].Name)) = 0 then
          begin
          Item.FileNode := Files[k];
          Break;
          end;
      end;
    end;
end;

procedure TCollectionBuilder.BuildConnections_FileList_Cache;
var F: TFileItem;
    Records, Node: TjanXmlNode2;
    Name, Path, TextTag: string;
    Title, TitleSort, Author, AuthorSort: string;
    NameLen, PathLen: integer;
    i, j: integer;
    bNew: boolean;
    D: TDateTime;
begin
  Records := GetCacheRecords;
  TextTag := GetTextTag;
  if Records = nil then
    begin
    Records := Cache.addChildByName('records');
    end;
  for i := 0 to Pred(Files.Count) do
    begin
    F := Files[i];
    F.XmlNode := nil;
    Name := LinuxPath(F.Name);
    NameLen := Length(Name);
    for j := 0 to Pred(Records.nodes.Count) do
      begin
      Node := Records.childNode[j];
      if Node.name = TextTag then
        begin
        Path := LinuxPath(Node.attribute['path']);
        PathLen := Length(Path);
        if ((PathLen = NameLen) and (AnsiCompareText(Path, Name) = 0)) or
           ((PathLen > NameLen) and (Path[PathLen-NameLen] = '/') and (AnsiCompareText(Copy(Path, PathLen-NameLen+1, NameLen), Name) = 0))
        then
          begin
          F.XmlNode := Node;
          Break;
          end;
        end;
      end;
    bNew := F.XmlNode = nil;
    if bNew then
      begin
      // Vytvorim novy zaznam do cache
      Node := Records.addChildByName(TextTag);
      Node.attribute['status'] := '0';
      Node.attribute['sourceid'] := '1';
      Node.attribute['id'] := NextId;
      Node.attribute['mime'] := FileNameToMimeType(F.Name);
      Node.attribute['path'] := LinuxPath(Config.BooksBasePath + F.Name);
      F.XmlNode := Node;
      end
    // Tohle fakticky neni potreba, protoze pokud v F.XmlNode neco je, tak je to Node,
    // ale co kdybych ten kod nekdy zmenil...
    else
      Node := F.XmlNode;
    if bNew or Config.RefreshExisting then
      begin
      if Config.MemoryType = mtInternal then
        D := F.CreationDate
      else
        D := F.Date;
      Node.attribute['date'] := DateTimeToXml(D + Config.TimeDelta);
      Node.attribute['size'] := F.Size;
      if Config.GenerateMetadata and GetMetadata(F.Directory + F.Name, Title, TitleSort, Author, AuthorSort) then
        begin
        Node.attribute['title'] := Title;
        Node.attribute['titleSorter'] := TitleSort;
        Node.attribute['author'] := Author;
        Node.attribute['authorSorter'] := AuthorSort;
        end;
      end;
    end;
end;

procedure TCollectionBuilder.BuildConnections_Collection_Cache;
var Col: TCollectionDef;
    ColTitle: string;
    Records, PlaylistNode, Node: TjanXmlNode2;
    PlaylistTag, PlaylistItemTag: string;
    i, j: integer;
    b: boolean;
begin
  Records := GetCacheRecords;
  PlaylistTag := GetPlaylistTag;
  PlaylistItemTag := GetPlaylistItemTag;
  for i := 0 to Pred(Collections.Count) do
    begin
    // Sbirka, kterou chci mit na ctecce
    Col := Collections[i];
    // Existuje v ni aspon jeden soubor?
    b := False;
    for j := 0 to Pred(Col.Count) do
      if Col[j].FileNode <> nil then
        begin
        b := True;
        Break;
        end;
    if b then
      begin
      ColTitle := Col.Title;
      if Config.ConvertToUTF then
        case Config.Encoding of
          enOEM:     ColTitle := OemToUTF8(ColTitle);
          enAnsi:    ColTitle := AnsiToUTF8(ColTitle);
          enUnicode: ColTitle := WideToUTF8(ColTitle);
          enUtf8:    ;
          end;
      // Potrebuju se podivat, jestli ta sbirka uz existuje
      PlaylistNode := nil;
      for j := 0 to Pred(Records.nodes.Count) do
        begin
        Node := Records.nodes[j];
        if Node.name = PlaylistTag then
          if Node.attribute['title'] = ColTitle then
            begin
            PlaylistNode := Records.nodes[j];
            Break;
            end;
        end;
      // Bud si sbirku zalozim, nebo ji jenom promazu
      if PlaylistNode = nil then
        begin
        PlaylistNode := Records.addChildByName(PlaylistTag);
        PlaylistNode.attribute['title'] := ColTitle;
        PlaylistNode.attribute['sourceid'] := 1;
        PlaylistNode.attribute['id'] := NextId;
        end
      else
        begin
        PlaylistNode.nodes.Clear;
        end;
      // Ted jednotlive existujici soubory priradim do playlistu
      for j := 0 to Pred(Col.Count) do
        if Col[j].FileNode <> nil then
          begin
          Node := PlaylistNode.addChildByName(PlaylistItemTag);
          Node.attribute['id'] := Col[j].FileNode.XmlNode.attribute['id'];
          end;
      end;
    end;
end;

function TCollectionBuilder.GetMetadata(const FN: string; out Title, TitleSort, Author, AuthorSort: string): boolean;
var Exe, CommandLine: string;
    Input, Output, Error: TStringList;
    ResCode: DWORD;
    i, j: integer;
    Name, Value: string;
    RE: TRegExpr;
begin
  Result := False;
  if Config.CalibrePath = '' then
    Exit;
  if not FileExists(FN) then
    Exit;
  Exe := '';
  case FileNameToFileType(FN) of
    ftLRF: Exe := Config.LRFMeta;
    ftPDF: Exe := Config.PDFMeta;
    ftRTF: Exe := Config.RTFMeta;
    ftTXT: Exe := Config.TXTMeta;
    ftEPUB: Exe := Config.EPUBMeta;
    end;
  if Exe = '' then
    Exit;
  if not IsAbsolutePath(Exe) then
    Exe := IncludeTrailingBackslash(Config.CalibrePath) + Exe;
  if not FileExists(Exe) then
    Exit;
  Input := TStringList.Create;
  try
    Output := TStringList.Create;
    try
      Error := TStringList.Create;
      try
        RE := nil;
        try
          CommandLine := Format('"%s" "%s"', [Exe, FN]);
          Writeln(CommandLine);
          if GetConsoleOutput(CommandLine, Input, Output, Error, ResCode, SW_HIDE) then
            if ResCode = 0 then
              begin
              Result := True;
              Title := '';
              TitleSort := '';
              Author := '';
              AuthorSort := '';
              for i := 0 to Pred(Output.Count) do
                if Output[i] <> '' then
                  begin
                  j := Pos(':', Output[i]);
                  if j > 0 then
                    begin
                    Name := UpperCase(Trim(Copy(Output[i], 1, Pred(j))));
                    Value := Trim(Copy(Output[i], Succ(j), MaxInt));
                    if Config.ConvertToUTF then
                      Value := AnsiToUTF8(Value); // It SHOULD be OemToUTF8, but Calibre incorrectly outputs in ANSI.
                    if Name = 'AUTHOR' then
                      Author := Value
                    else if Name = 'AUTHOR.READING' then
                      AuthorSort := Value
                    else if Name = 'TITLE' then
                      Title := Value
                    else if Name = 'TITLE.READING' then
                      TitleSort := Value
                    // Calibre 0.6
                    else if Name = 'AUTHOR(S)' then
                      begin
                      if RE = nil then
                        begin
                        RE := TRegExpr.Create;
                        RE.ModifierI := True;
                        RE.Expression := '^(.*?)\s*\[(.*)\]$';
                        end;
                      if RE.Exec(Value) then
                        begin
                        Author := RE.Match[1];
                        AuthorSort := RE.Match[2];
                        end
                      else
                        Author := Value;
                      end
                    else if Name = 'TITLE SORT' then
                      TitleSort := Value
                      ;
                    end;
                  end;
              end;
        finally
          if RE <> nil then
            RE.Free;
          end;
      finally
        Error.Free;
        end;
    finally
      Output.Free;
      end;
  finally
    Input.Free;
    end;
end;

procedure TCollectionBuilder.Execute;
var PlayListPath, CacheIn, BackupFN: string;
    bPlayList: boolean;
begin
  bPlayList := Config.PlaylistDefinitionFileName <> '';
  if bPlayList then
    PlayListPath := ExtractFilePath(Config.PlaylistDefinitionFileName)
  else
    PlayListPath := '';
  if bPlayList then
    Collections.LoadFromFile(Config.PlaylistDefinitionFileName);
  if Config.CacheInFileName = '' then
    Raise ECollectionBuilder.Create('Input cache file not defined.');
  if IsAbsolutePath(Config.CacheInFileName) then
    CacheIn := Config.CacheInFileName
  else
    CacheIn := PlayListPath + Config.CacheInFileName;
  {$IFDEF ALLOW_NONEXISTENT_CACHE}
    if FileExists(CacheIn) then
      Cache.LoadXML(CacheIn)
    else if Config.MemoryType in [mtMS, mtSD] then
      Cache.Xml := DEFAULT_XML_SD
    else
      Cache.Xml := DEFAULT_XML_INTERNAL;
  {$ELSE}
    Cache.LoadXML(CacheIn);
  {$ENDIF}
  if Config.BooksPath = '' then
    Config.BooksPath := '.\'
  else
    Config.BooksPath := IncludeTrailingBackslash(Config.BooksPath);
  Files.LoadFromDir(Config.BooksPath, Config.Recursive, Config.FileNameRegExp);
  BuildConnections;
  if Config.CacheOutFileName = '' then
    begin
    BackupFN := CacheIn + '.bak';
    if FileExists(BackupFN) then
      SysUtils.DeleteFile(BackupFN);
    CopyFile(PChar(CacheIn), PChar(BackupFN), FALSE);
    Config.CacheOutFileName := CacheIn;
    end;
  if IsAbsolutePath(Config.CacheOutFileName) then
    Cache.SaveXML(Config.CacheOutFileName)
  else
    Cache.SaveXML(PlayListPath + Config.CacheInFileName);
end;

end.
